//tcpserver.java Q1 2004
//Features:
// TCP server listen socket
// TCP stream (send/receive)
// Multiple connection management

import java.io.*;
import java.net.*;
import java.util.Hashtable;
import java.util.Enumeration;

public class tcpserver extends Thread {
	static final int version = 5;

	//Modules
	static loggers    lm;
	static config     cfg;
	static billerprot core;

	//TCP
	static ServerSocket billersocket;
	static Listener     listener;
	
	static Keepalive    keepalive;
	static Hashtable    servers;
	static boolean      bKeepGoing = false;

//// Send ////

	public void sendRaw(String port, String str)
	{
		Server s = null;
	
		//Loop all servers
		Enumeration e = servers.elements();
		while(e.hasMoreElements())
		{
			s = (Server)e.nextElement();
			if (port.equals(s.serverid))
			{
				s.send.println(str);
				//lm.Log("D", "<tcpserver> Sending: " + str);
				return;
			}
		}
	}

	public void SendAllRaw(String str)
	{
		Server s = null;
	
		//Loop all servers
		Enumeration e = servers.elements();
		while(e.hasMoreElements())
		{
			s = (Server)e.nextElement();
			if (s != null)
			{
				s.send.println(str);
				//lm.Log("D", "<tcpserver> Sending: " + str);
			}
		}
	}

	public void SendSoundMessage(String port, String pid, String sound,
		String text)
	{
		//MSG:pid:sound:text
		sendRaw(port, "MSG:" + pid + ":" + sound + ":" + text);
	}

	public void SendMessage(String port, String pid, String text)
	{
		//MSG:pid:sound:text
		//sendRaw(port, "MSG:" + pid + ":0:" + text);
		SendSoundMessage(port, pid, "0", text);
	}

	public void SendStaffSoundMessage(String port, String sender,
		String sound, String text)
	{
		//STAFFMSG:sender:sound:text
		sendRaw(port, "STAFFMSG:" + sender + ":" + sound + ":" + text);
	}

	public void SendStaffSoundMessage(String port, String sound,
		String text)
	{
		//STAFFMSG:sender:sound:text
		//sendRaw(port, "STAFFMSG::" + sound + ":" + text);
		SendStaffSoundMessage(port, "", sound, text);
	}


	public void SendStaffMessage(String port, String sender, String text)
	{
		//STAFFMSG:sender:sound:text
		//sendRaw(port, "STAFFMSG:" + sender + ":0:" + text);
		SendStaffSoundMessage(port, "0", sender, text);
	}

	public void SendStaffMessage(String port, String text)
	{
		//STAFFMSG:sender:sound:text
		//sendRaw(port, "STAFFMSG::0:" + text);
		SendStaffMessage(port, "", text);
	}


	public void SendBroadcast(String sender, String sound, String text)
	{
		//BROADCAST:sender:sound:text
		SendAllRaw("BROADCAST:" + sender + ":" + sound + ":" + text);
	}

	public void SendBroadcast(String sound, String text)
	{
		//BROADCAST:sender:sound:text
		//sendRaw("BROADCAST::" + sound + ":" + text);
		SendBroadcast("", sound, text);
	}

	public void SendBroadcast(String text)
	{
		//BROADCAST:sender:sound:text
		//sendRaw("BROADCAST::0:" + text);
		SendBroadcast("0", text);
	}

//// Core ////

	public tcpserver(loggers lm, config cfg, billerprot core)
	{	
		this.lm   = lm;
		this.cfg  = cfg;
		this.core = core;

		servers = new Hashtable();

		try
		{
			int port = cfg.GetInt("Billing", "Port", 1850);
			//lm.Log("I", "<tcpserver> Listening on port " + port);
			billersocket = new ServerSocket(port, 2); //backlog=2 anti DOS
		}
		catch (Exception e)
		{
			lm.Log("E", "<tcpserver> " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
	}

	public void listen()
	{
		listener = new Listener();
		listener.start();
		
		keepalive = new Keepalive();
		keepalive.start();
		
		bKeepGoing = true;
		this.start();
	}

	private void receive()
	{
		Server s = null;
	
		//Loop all servers
		Enumeration e = servers.elements();
		while(e.hasMoreElements())
		{
			s = (Server)e.nextElement();
			receive(s);
		}
	}

	private void receive(Server s)
	{
		if (s == null) return;
	
		try
		{
			//if (s.recv.ready())
			{
				core.mainHandler(s.serverid, s.recv.readLine());
			}/*
			else if (!s.socket.isConnected() || !s.socket.isBound() ||
				s.socket.isClosed() || s.socket.isInputShutdown() ||
				s.socket.isOutputShutdown())
			{
				disconnect(s);
				lm.Log("D", "<tcpserver> [" + s.serverid +
					"] disconnected");
			}*/
		}
		catch(NullPointerException e)
		{
			data pd = core.getDataRef(); //hack
			String name = pd.getZoneByPort(s.serverid);
			lm.Log("I", "<tcpserver> [" + name + "] disconnected");
			lm.Log("I", "<tcpserver> [" + name + "] Logging out " +
				"connected players...");
			pd.setAllOffline(s.serverid);
			disconnect(s);
		}
		catch(SocketTimeoutException e)
		{
			/* Ignore */
		}
		catch(SocketException e)
		{
			data pd = core.getDataRef(); //hack
			String name = pd.getZoneByPort(s.serverid);
			lm.Log("I", "<tcpserver> [" + name + "] disconnected");
			lm.Log("I", "<tcpserver> [" + name + "] Logging out " +
				"connected players...");
			pd.setAllOffline(s.serverid);
			disconnect(s);
		}
		catch(Exception e)
		{
			lm.Log("E", "<tcpserver> [" + s.serverid + "] " +
				"Receive error: " + e.getMessage());
			e.printStackTrace();
		}
	}

	public void run()
	{
		while(bKeepGoing)
		{
			try
			{
				Thread.sleep( 200 ); //0.2 second
			}
			catch(InterruptedException e)
			{
				lm.Log("E", "<tcpserver> " +
					e.getMessage());
			}
			catch(Exception e)
			{
				lm.Log("E", "<tcpserver> " +
					e.getMessage());
				e.printStackTrace();
			}

			if (servers.size() > 0)
			{
				receive();
			}
		}
	}

	//Disconnect all, including listen socket
	public void disconnect()
	{
		Server s = null;
	
		//Loop all servers
		Enumeration all = servers.elements();
		while(all.hasMoreElements())
		{
			s = (Server)all.nextElement();
			disconnect(s);
		}
		
		lm.Log("I", "<tcpserver> WARNING: Disconnected from all servers");

		listener.halt();
		keepalive.halt();
		
		try
		{
			billersocket.close();
		}
		catch(Exception e)
		{
			lm.Log("E", "<tcpserver> " + e.getMessage());
			e.printStackTrace();
			return;
		}
	}

	public void disconnect(String port)
	{
		Server s = null;
	
		//Loop all servers
		Enumeration all = servers.elements();
		while(all.hasMoreElements())
		{
			s = (Server)all.nextElement();
			if (port.equals(s.serverid))
				disconnect(s);
		}
	}

	public void disconnect(Server s)
	{
		if (s == null) return;
		servers.remove(s.serverid);
	
		try
		{
			s.send.close();
			s.recv.close();
			s.socket.close();
		}
		catch(NullPointerException e)
		{
			lm.Log("E", "<tcpserver> " + e.getMessage());
			e.printStackTrace();
			//If it couldn't make the socket then send, recv
			// and socket are probably null
			return;
		}
		catch(Exception e)
		{
			lm.Log("E", "<tcpserver> " + e.getMessage());
			e.printStackTrace();
			return;
		}
	}

//// Blocking stuff is put in separate thread ////

	class Listener extends Thread
	{
		protected boolean bKeepListening = false;

		public Listener()
		{
			bKeepListening = true;
		}

		public void run()
		{
			Server s      = null;
			Socket socket = null;
			lm.Log("I", "<tcpserver> Listening on port " +
				billersocket.getLocalPort());
			
			while(bKeepListening)
			{
				try
				{
					Thread.sleep( 2000 ); //2 seconds
				}
				catch(InterruptedException e)
				{
					lm.Log("E", "<tcpserver> " +
						e.getMessage());
				}
				catch(Exception e)
				{
					lm.Log("E", "<tcpserver> " +
						e.getMessage());
					e.printStackTrace();
				}
			
				try {	
					socket = billersocket.accept();
					socket.setKeepAlive(true);
					socket.setSoTimeout(200);
					//System.out.println("timeout: " + socket.getSoTimeout());
					//System.out.println(socket.getSoLinger());
					//System.out.println(socket.getKeepAlive());
					//System.out.println(socket.getSendBufferSize());
					//System.out.println(socket.getReceiveBufferSize());
					s = new Server(socket);
						lm.Log("I", "<tcpserver> new " +
							"connection from " +
							s.serverid);

					s.send = new PrintWriter(
						socket.getOutputStream(), true);
					s.recv = new BufferedReader(
						new InputStreamReader(
						socket.getInputStream()));

					servers.put(s.serverid, s);
				}
				catch (IOException e)
				{
					lm.Log("D", "<tcpserver> listen: " +
						e.getMessage());
					disconnect(s);
					return;
				}
				catch (Exception e)
				{
					lm.Log("E", "<tcpserver> listen: " +
						e.getMessage());
					e.printStackTrace();
					disconnect(s);
				}
			}
		}
		
		public void halt()
		{
			bKeepListening = false;
			this.interrupt();
		}
	}

	class Keepalive extends Thread
	{
		protected boolean bKeepAlive = false;

		public Keepalive()
		{
			bKeepAlive = true;
		}

		public void run()
		{
			while(bKeepAlive)
			{
				try
				{
					Thread.sleep( 180 * 1000 ); //180 seconds
				}
				catch(InterruptedException e)
				{
//					lm.Log("E", "<tcpserver> keepalive: " +
//						e.getMessage());
				}
			
				SendAllRaw("PING");
			}
		}
		
		public void halt()
		{
			bKeepAlive = false;
			this.interrupt();
		}
	}

//// Containers ////

	class Server
	{
		static final int version = 1;
	
		Socket       socket;
		PrintWriter    send;
		BufferedReader recv;
		
		String     serverid;
		
		public Server(Socket socket)
		{
			this.socket = socket;
/*
			send = new PrintWriter(
				socket.getOutputStream(), true);
			recv = new BufferedReader(
				new InputStreamReader(
				socket.getInputStream()));
*/
			serverid = socket.getInetAddress().getHostAddress() +
				":" + socket.getPort();
		}
	}
}
